/**
 * Copyright (C) 2003 France Telecom R&D
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
package org.objectweb.perseus.persistence.concurrency;

import java.util.Iterator;

import org.objectweb.perseus.concurrency.optimistic.OptimisticConcurrencyManager;
import org.objectweb.perseus.concurrency.lib.TimeStamp;
import org.objectweb.perseus.concurrency.api.ConcurrencyException;
import org.objectweb.perseus.persistence.api.*;
import org.objectweb.perseus.cache.api.CacheEntry;
import org.objectweb.util.monolog.api.BasicLevel;

/**
 *
 * @author S.Chassande-Barrioz
 */
public class POptimisticConcurrencyManager extends OptimisticConcurrencyManager {


    public final static String STORAGE_MANAGER_BINDING = "storage-manager";
    public final static String STATE_MANAGER_BINDING = "state-manager";

    protected StateManager stateManager;

    protected StorageManager storageManager;

    //IMPLEMENTATION OF THE UserBindingControler INTERFACE //
    //------------------------------------------------------//

    public String[] listFc() {
		String[] super_res = super.listFc();
		String res[] = new String[super_res.length + 2];
		for (int cpt = 0; cpt < super_res.length; cpt++) {
			res[cpt] = super_res[cpt];
		}
        res[super_res.length + 0] = STORAGE_MANAGER_BINDING;
		res[super_res.length + 1] = STATE_MANAGER_BINDING;
		return res;
    }

    public Object lookupFc(String s) {
        if (STORAGE_MANAGER_BINDING.equals(s)) {
            return storageManager;
        } else if (STATE_MANAGER_BINDING.equals(s)) {
            return stateManager;
        } else {
            return super.lookupFc(s);
        }
    }

    public void bindFc(String s, Object o) {
        if (STORAGE_MANAGER_BINDING.equals(s)) {
            storageManager = (StorageManager) o;
        } else if (STATE_MANAGER_BINDING.equals(s)) {
            stateManager = (StateManager) o;
        } else super.bindFc(s, o);
    }

    public void unbindFc(String s) {
        if (STORAGE_MANAGER_BINDING.equals(s)) {
            storageManager = null;
        } else if (STATE_MANAGER_BINDING.equals(s)) {
            stateManager = null;
        } else {
            super.unbindFc(s);
        }
    }
    public void finalize(Object arg0) {
        WorkingSet ws = (WorkingSet) arg0;
        boolean retainValues = ws.getWSRetainValues();
		for (Iterator it = ws.entries().iterator(); it.hasNext();) {
			State state = (State) it.next();
			if (state == VirtualState.instance) {
				continue;
			}
			CacheEntry ce = state.getCacheEntry();
			if (!stateManager.isUnexported(state)) {
			    //The object has not been unexported (deleted)
				if (retainValues) {
				    if (stateManager.isDirty(state)) {
				        if (stateManager.isToMerge(state)) {
				            synchronized(ce) {
				                stateManager.setReferenceState(ce,
					                stateManager.merge(
					                        stateManager.getReferenceState(ce),
					                        state));
				            }
				        } else {
				            stateManager.makeClean(state);
				            stateManager.setReferenceState(ce, state);
				        }
				    }//else state is clean ==> nothing to do
				} else if (stateManager.getReferenceState(ce) == state) {
				    //forget the state used as reference state
					stateManager.setReferenceState(ce, null);
				}
			} else {
				// Assign the current state as reference to the 
				// unexported object
				stateManager.setReferenceState(ce, state);
			}
		}
        super.finalize(arg0);
    }

	public void abort(Object ctx) {
	    boolean debug = logger != null && logger.isLoggable(BasicLevel.DEBUG);
	    if (debug) {
	        logger.log(BasicLevel.DEBUG, "Abort the context: " + ctx);
	    }
	    WorkingSet ws = (WorkingSet) ctx;
	    for(Iterator it = ws.entries().iterator(); it.hasNext();) {
	        State state = (State) it.next();
	        if (state == VirtualState.instance) {
	        	continue;
	        }
			CacheEntry ce = state.getCacheEntry();
			if (stateManager.isUnexported(state)) {
				stateManager.makeClean(state);
				stateManager.makeBound(ce, ce.getCeIdentifier());
			} else if (stateManager.isExported(state)) {
				stateManager.makeClean(state);
			}
	    }
	    super.abort(ctx);
	}

    public void closeTimeStamp(TimeStamp ts, Object ctx) {
		WorkingSet ws = (WorkingSet) ctx;
		State s = ws.lookup(ts.oid);
		if (s != null && s != VirtualState.instance) {
		    stateManager.stateNoMoreUsed(s);
		}
    }
	protected Object getState(Object ctx,
							  Object resource,
							  TimeStamp timeStamp,
							  boolean isDirtyBefore,
							  boolean isWrite) throws ConcurrencyException {
		CacheEntry ce = (CacheEntry) resource;
		WorkingSet ws = (WorkingSet) ctx;
		State state = ws.lookup(ce.getCeIdentifier());
		if (state == VirtualState.instance) {
		    state = null;
		}
		if (state == null) {
			state = stateManager.getReferenceState(ce);
			if (state == null) { //no reference state ==> load it
				state = load(ws, ce);
			}
			if (isWrite) {
				//copy on write
				state = stateManager.createState(state);
			} // else  read case
		} else {
			//There is state in the working set
			if (isWrite) {
				if (!isDirtyBefore) {
					//copy on write
					state = stateManager.createState(state);
				} //else already in write
			} // else read case ==> use the state of the workingset
		}
		return state;
    }

	private State load(WorkingSet ws, CacheEntry ce)
		throws ConcurrencyException {
		State state = stateManager.createState(ce);
		stateManager.setReferenceState(state.getCacheEntry(), state);
		try {
			storageManager.read(ws, ce.getCeIdentifier(), state);
		} catch (PersistenceException e) {
			stateManager.setReferenceState(state.getCacheEntry(), null);
			stateManager.destroyState(state);
			if (e instanceof NoDSIPersistenceException) {
				throw new NoDSIConcurrencyException(e);
			} else {
			    throw new ConcurrencyException(e);
			}
		}
		stateManager.makeClean(state);
		return state;
	}
	
	protected Object getResourceId(Object object) {
	    return ((CacheEntry) object).getCeIdentifier();
	}

}
